{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "c74fe0f0",
   "metadata": {},
   "source": [
    "# Module: gd071\n",
    "\n",
    "\n",
    "* Author: Johannes Busse, https://www.jbusse.de/, jbusse at jbusse dot de\n",
    "* Homepage of GenDifS: https://www.jbusse.de/gendifs\n",
    "* Version: gd071, 2024-10-23\n",
    "\n",
    "License: CC BY-NC 4.0 [DE](https://creativecommons.org/licenses/by-nc/4.0/legalcode.de) | [EN](https://creativecommons.org/licenses/by-nc/4.0/legalcode.en)\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "fd7214cf",
   "metadata": {
    "lines_to_next_cell": 0
   },
   "outputs": [],
   "source": [
    "from lxml import etree # https://lxml.de/\n",
    "import rdflib\n",
    "import owlrl\n",
    "\n",
    "# global variables\n",
    "verbose = None\n",
    "NewID = 1000"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1eb0d1de",
   "metadata": {},
   "outputs": [],
   "source": [
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "925c9ef2",
   "metadata": {},
   "source": [
    "## GenDifS\n",
    "\n",
    "main class."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5d0ccc67",
   "metadata": {},
   "outputs": [],
   "source": [
    "class GenDifS:\n",
    "    \"\"\"Main class managing file info; contains dict of GenDifS_taxonomy classes\"\"\"\n",
    "\n",
    "    def __init__(self, mindmap_file_location=None, *, from_string=None, verb = 1):\n",
    "\n",
    "        global verbose\n",
    "        verbose = verb\n",
    "\n",
    "        if mindmap_file_location != None:\n",
    "            # open .mm file, parse xml\n",
    "            xml_parser = etree.XMLParser(remove_blank_text=True)\n",
    "            self.mindmap_file_location = mindmap_file_location\n",
    "            self.mindmap_xml = etree.parse(self.mindmap_file_location, xml_parser)\n",
    "        elif from_string != None:\n",
    "            self.mindmap_xml = self.xml_mindmap_from_string(from_string)\n",
    "        else:\n",
    "            print(\"ERROR 53: no mindmap given, nothing to do\")\n",
    "            return\n",
    "\n",
    "        self.mindmap_topnode = self.mindmap_xml.getroot().find(\"node\") # upmost xml node with xml name \"node\"\n",
    "\n",
    "        \n",
    "        \n",
    "        self.dict_of_all_taxonomy_xlm_nodes = None\n",
    "        self.taxonomies_by_ID = None\n",
    "        self.taxonomies_by_name = None\n",
    "        self.code = None\n",
    "        self.languages = None\n",
    "\n",
    "        \n",
    "    def xml_mindmap_from_string(self, string):\n",
    "\n",
    "        def depth_lstrip(x):\n",
    "            x = x.rstrip()\n",
    "            xlstrip = x.lstrip()\n",
    "            depth = len(x) - len(xlstrip)\n",
    "            return depth, xlstrip\n",
    "\n",
    "        def texttree_to_xml(current_depth, parent):\n",
    "            nonlocal i, texttree, texttree_len\n",
    "\n",
    "            #subtree = []\n",
    "            new_parent = parent\n",
    "\n",
    "            while i < texttree_len:\n",
    "                d, txt = depth_lstrip(texttree[i])\n",
    "                #print(\"10:\", d, txt)\n",
    "                if d == current_depth:\n",
    "                    if len(txt) > 0:\n",
    "                        #subtree.append(txt)\n",
    "                        child = etree.SubElement(parent, \"node\")\n",
    "                        child.set(\"TEXT\", txt)\n",
    "\n",
    "                        dct = string2dict(txt)\n",
    "                        if \"ID\" in dct.keys():\n",
    "                            child.set(\"ID\", \"_\".join(dct[\"ID\"]))\n",
    "                        else:\n",
    "                            child.set(\"ID\", f\"{i}\")\n",
    "                        new_parent = child\n",
    "                    i += 1\n",
    "                elif d > current_depth:\n",
    "                    texttree_to_xml(d, new_parent)\n",
    "                else:\n",
    "                    return parent\n",
    "            return parent\n",
    "    \n",
    "        i = 0\n",
    "        texttree = [\"xml_mindmap_from_string\"]\n",
    "        texttree.extend( [ '  ' + s for s in string.split(\"\\n\") ] )\n",
    "        texttree_len = len(texttree)\n",
    "        root_map = etree.Element(\"map\")\n",
    "        #node_one = etree.SubElement(root_map, \"node\")\n",
    "        #node_one.set(\"TEXT\", \"XML mm from string\")\n",
    "        #root_et = etree.ElementTree(root_map)\n",
    "        xml = texttree_to_xml(i, root_map)\n",
    "        return etree.ElementTree(xml)\n",
    "\n",
    "\n",
    "    \n",
    "    def compile(self): # , language_list_list = [] \n",
    "\n",
    "        # self.language_list_list = language_list_list\n",
    "        \n",
    "        # create list of TAXONOMY xml nodes\n",
    "        self.dict_of_all_taxonomy_xlm_nodes = {}\n",
    "        self.search_mm_for_taxonomy_xml_nodes(self.mindmap_topnode, self.dict_of_all_taxonomy_xlm_nodes)\n",
    "        if verbose >= 3:\n",
    "            print(\"INFO 131\", self.dict_of_all_taxonomy_xlm_nodes)\n",
    "    \n",
    "        # for each TAXONOMY node: allocate a GenDifS_Taxonomy object \n",
    "        self.taxonomies_by_ID = {}\n",
    "        self.taxonomies_by_name = {}\n",
    "        for xml_node_ID, xml_node in self.dict_of_all_taxonomy_xlm_nodes.items():\n",
    "\n",
    "            # create taxonomy\n",
    "            t = GenDifS_Taxonomy(xml_node_ID, xml_node ) # , language_list_list = self.language_list_list\n",
    "            \n",
    "            # taxonomies by ID\n",
    "            self.taxonomies_by_ID[xml_node_ID] = t\n",
    "            \n",
    "            # taxonomies by name\n",
    "            # t.tree.context is the root node of each taxonomy tree, i.e. the node with \"TAXONOMY\"\n",
    "            t_name = t.tree.context[\"taxonomy_name\"]\n",
    "            if t_name in self.taxonomies_by_name:\n",
    "                print(f\"WARNING 62: multiple use of taxonomy name {t_name}; using ID {xml_node_ID} instead.\")\n",
    "                t_name = xml_node_ID\n",
    "            self.taxonomies_by_name[t_name] = t\n",
    "\n",
    "            \n",
    "    def describe_mindmap(self):\n",
    "        \"\"\"Mindmap metadata: #nodes etc.\"\"\"\n",
    "        print(f\"{self.mindmap_topnode.get('TEXT')=}\")\n",
    "        print(f\"{len(self.mindmap_xml.getroot().xpath('.//node'))=} nodes\")\n",
    "\n",
    "    # def describe_taxonomies(self, indent=True):\n",
    "    #     for t in self.taxonomies_by_ID.values():\n",
    "    #         t.describe()\n",
    "\n",
    "    # def content_taxonomies(self, indent=True):\n",
    "    #     for t in self.taxonomies_by_ID.values():\n",
    "    #         t.describe_contexts(indent)\n",
    "\n",
    "    # Identify the xml nodes witch define a taxonomy.\n",
    "    # Each taxonomy will be translated into an own standalone named rdf graph.\n",
    "    # All taxonomies together will contribute to a rdf dataset.\n",
    "    # The default graph contains the metadata of the whole ensemble.\n",
    "    \n",
    "    def search_mm_for_taxonomy_xml_nodes(self, xml_node, dict_of_all_taxonomy_nodes):\n",
    "        global NewID\n",
    "        my_text = xml_node.get('TEXT')\n",
    "        if my_text == \"TAXONOMY\" or my_text.startswith(\"TAXONOMY \"):\n",
    "            # print(f\"search_gdntree_for_taxonomy: {xml_node.get('TEXT')=}\")\n",
    "                \n",
    "            # it's essential to have unique node IDs\n",
    "            ID = xml_node.get('ID', None)\n",
    "            if ID == None or ID in dict_of_all_taxonomy_nodes:\n",
    "                NewID += 1\n",
    "                ID = f\"NewID_{NewID}\"\n",
    "                xml_node[\"ID\"] = ID\n",
    "            dict_of_all_taxonomy_nodes[ID] = xml_node\n",
    "    \n",
    "        # we allow for nested TAXONOMY trees\n",
    "        for child_xml_node in xml_node.findall('node'):\n",
    "            button_cancel = 'button_cancel' in [ icon.attrib['BUILTIN'] for icon in child_xml_node.findall(\"icon\") ]\n",
    "            if not button_cancel:\n",
    "                self.search_mm_for_taxonomy_xml_nodes(child_xml_node, dict_of_all_taxonomy_nodes)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "41140e8c",
   "metadata": {},
   "outputs": [],
   "source": [
    "class GenDifS_Taxonomy():\n",
    "    \"\"\"class managing a stand alone taxonomy; contains root element of a tree of gdn objects\"\"\"\n",
    "\n",
    "    def __init__(self, xml_ID, xml_node): # , language_list_list = []\n",
    "        self.ID = xml_ID\n",
    "        self.TEXT = xml_node.get(\"TEXT\")\n",
    "\n",
    "        # TBD: generate terminology specific namespaces?\n",
    "        self.default_namespaces_dict = {\n",
    "            \"ex\": \"<http://example.net/namespace/ex#>\" ,  # examples of OWL A-Box\n",
    "            \"cpt\": \"<http://example.net/namespace/cpt#>\" ,    # concept; instance of skos:concept\n",
    "            \"sheet\": \"<http://example.net/namespace/sheet#>\" ,    # instances from Excel-Sheet\n",
    "            \"\": \"<http://example.net/namespace/default#>\" , # OWL T-Box\n",
    "            \"owl\": \"<http://www.w3.org/2002/07/owl#>\" ,\n",
    "            \"rdf\": \"<http://www.w3.org/1999/02/22-rdf-syntax-ns#>\" ,\n",
    "            \"xml\": \"<http://www.w3.org/XML/1998/namespace>\" ,\n",
    "            \"xsd\": \"<http://www.w3.org/2001/XMLSchema#>\" ,\n",
    "            \"rdfs\": \"<http://www.w3.org/2000/01/rdf-schema#>\" ,\n",
    "            \"skos\": \"<http://www.w3.org/2004/02/skos/core#>\" ,\n",
    "            \"gendifs\": \"<http://jbusse.de/gendifs#>\" ,\n",
    "        }\n",
    "\n",
    "        ttl_prefixes =    lambda dct: \"\\n\".join([f\"@prefix {key}: {val} .\" for key, val in dct.items()])\n",
    "        sparql_prefixes = lambda dct: \"\\n\".join([f\"PREFIX  {key}: {val}  \" for key, val in dct.items()])            \n",
    "        self.ttl_namespaces = ttl_prefixes(self.default_namespaces_dict)\n",
    "        self.sparql_namespaces = sparql_prefixes(self.default_namespaces_dict)\n",
    "        # print(f\"INFO 238:{self.ttl_namespaces=}\")\n",
    "        \n",
    "        # Dictionary { ID: GenDifS Node } of all GenDifS nodes of the current taxonomy. \n",
    "        self.dict_of_all_gdn = {}\n",
    "\n",
    "        \n",
    "        # \n",
    "        # walk the tree several times, add more and more information\n",
    "        #\n",
    "\n",
    "        # Step 1:\n",
    "        # build tree of gdn (GenDifS nodes)\n",
    "        self.tree = self.build_gdntree_from_mm(xml_node, None, 0, self.dict_of_all_gdn )\n",
    "        # self.dict_of_all_gdn now is a dict of all gdn (GenDifS node) objects\n",
    "\n",
    "        # Step 2:\n",
    "        # collect context information of each gdn to prepare context free code generation\n",
    "        # [  for gdn in self.dict_of_all_gdn.values() ]\n",
    "        for gdn in self.dict_of_all_gdn.values():\n",
    "            gdn.collect_context()\n",
    "            \n",
    "        # Step 3:\n",
    "        # generate ttl code and integrity information\n",
    "        # [ gdn.generate_molecule_per_language() for gdn in self.dict_of_all_gdn.values() ]\n",
    "        for gdn in self.dict_of_all_gdn.values():\n",
    "            gdn.compile()\n",
    "\n",
    "        # Step 4:\n",
    "        # collect ttl code per language and allocate respective rdflib graphs\n",
    "        raw_ttl_lists = {}\n",
    "        self.rdf_graphs = {}\n",
    "        for gdn_id, gdn_val in self.dict_of_all_gdn.items():\n",
    "            for code_lang, code_val in gdn_val.code.items():\n",
    "                if code_lang not in raw_ttl_lists:\n",
    "                    raw_ttl_lists[ code_lang ] = [ self.ttl_namespaces ]\n",
    "                    self.rdf_graphs[ code_lang ] = rdflib.Graph()\n",
    "                raw_ttl_lists[ code_lang ].append( code_val )\n",
    "                try:\n",
    "                    self.rdf_graphs[ code_lang ].parse(data= self.ttl_namespaces + code_val)\n",
    "                except:\n",
    "                    print(f\"ERROR 220: RDFLIB complains about\")\n",
    "                    print(code_val, \"\\n\")\n",
    "        self.raw_ttl = { k: \"\\n\\n\".join(v) for k,v in raw_ttl_lists.items() }\n",
    "            \n",
    "            \n",
    "    def build_gdntree_from_mm(self, xml_node, parent_gdn, depth, dict_of_all_gdn):\n",
    "        \"\"\"build tree of gdn (GenDifS nodes); store result in self.dict_of_all_gdn\"\"\"\n",
    "        \n",
    "        # allocate new instance of gdn (gdn = GenDifS Node)\n",
    "        if parent_gdn == None: # we are called from __init__\n",
    "            this_gdn = GenDifS_Differentia(xml_node, depth=depth, parent_gdn=None )\n",
    "        elif isinstance(parent_gdn, GenDifS_Differentia):\n",
    "            this_gdn = GenDifS_Taxon(xml_node, depth=depth, parent_gdn=parent_gdn )\n",
    "        elif isinstance(parent_gdn, GenDifS_Taxon):  # NEU 2024-09-19\n",
    "            if xml_node.get('TEXT', '').startswith('ATT'):\n",
    "                this_gdn = gdn_attribute(xml_node, depth=depth, parent_gdn=parent_gdn ) # 2024-09-19: TBD\n",
    "            else:\n",
    "                this_gdn = GenDifS_Differentia(xml_node, depth=depth, parent_gdn=parent_gdn )\n",
    "        else:\n",
    "            print(\"ERROR 294: 'else' should not be selected\")\n",
    "            \n",
    "        # add current gdn to dict_of_all_gdn\n",
    "        dict_of_all_gdn[this_gdn.id] = this_gdn\n",
    "        \n",
    "        for child_xml_node in xml_node.findall('node'):\n",
    "            button_cancel = 'button_cancel' in [ icon.attrib['BUILTIN'] for icon in child_xml_node.findall(\"icon\") ]\n",
    "            taxonomy_node = child_xml_node.get('TEXT', '').startswith('TAXONOMY')\n",
    "\n",
    "            if not ( button_cancel or taxonomy_node ):\n",
    "                this_gdn.children_gdn_list.append( self.build_gdntree_from_mm(child_xml_node, this_gdn, depth+1, dict_of_all_gdn) )\n",
    "        return this_gdn\n",
    "\n",
    "\n",
    "    def DEPRECATED_augment(self, abox):\n",
    "\n",
    "        \n",
    "        def instances(graph, category):\n",
    "            #q = ns_sparql + f\"\"\"SELECT ?i\n",
    "            q = f\"\"\"SELECT ?i\n",
    "            WHERE {{\n",
    "                ?i a {category} .\n",
    "            }}\"\"\"\n",
    "            return { r[0].encode() for r in graph.query(q) }\n",
    "\n",
    "        \n",
    "        if not \"RDFS_OWL\" in self.rdf_graphs:\n",
    "            print(f\"249, {self=}: 'RDFS_OWL' not in {self.rdf_graphs=}\")\n",
    "            return None\n",
    "\n",
    "        self.g_augment_before = rdflib.Graph()\n",
    "        self.g_augment_before.parse( data=self.rdf_graphs[\"RDFS_OWL\"].serialize() )\n",
    "        try:\n",
    "            self.g_augment_before.parse(data=abox)\n",
    "        except:\n",
    "            if verbose >= 1:\n",
    "                print(f\"WARNING 258: RDFLIB complains about abox, augmentation failed\")\n",
    "                return\n",
    "            \n",
    "        if verbose >= 2:\n",
    "            print(f\"augment() 264: {len(self.g_augment_before)=}\", end=\", \")\n",
    "\n",
    "        self.g_augment = rdflib.Graph()\n",
    "        self.g_augment.parse( data=self.g_augment_before.serialize() )\n",
    "        \n",
    "        owlrl.DeductiveClosure(owlrl.OWLRL_Semantics, axiomatic_triples = False).expand(self.g_augment)\n",
    "        if verbose >= 2:\n",
    "            print(f\"{len(self.g_augment)=}\")\n",
    "        \n",
    "        for gdn in self.dict_of_all_gdn.values():\n",
    "            #print(f\"261: {gdn.entity_context}\")\n",
    "            gdn.augment_genus_species_DEPRECATED(self.g_augment)\n",
    "            \n",
    "    def describe(self):\n",
    "        # print(f\"{self.ID} {self.TEXT}: #{len(self.dict_of_all_gdn)} GenDifS nodes;\\n{self.languages=};\\n{self.rdf_graphs.keys()=}\")\n",
    "        print(f\"{self.ID} {self.TEXT}: #{len(self.dict_of_all_gdn)} GenDifS nodes\")\n",
    "\n",
    "    # def describe_contexts(self, indent=True):\n",
    "    #     \"\"\"walk tree and describe nodes\"\"\"\n",
    "\n",
    "    #     def describe_contexts_walk(this_gdn, indent):  \n",
    "    #         this_gdn.describe(indent=indent)\n",
    "    #         for child_gdn in this_gdn.children_gdn_list:\n",
    "    #             describe_contexts_walk(child_gdn,indent)\n",
    "\n",
    "    #     describe_contexts_walk(self.tree, indent)\n",
    "\n",
    "\n",
    "\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e1a4b854",
   "metadata": {
    "lines_to_next_cell": 0
   },
   "outputs": [],
   "source": [
    "# ## gdn"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ae4735a7",
   "metadata": {},
   "outputs": [],
   "source": [
    "\n",
    "\n",
    "class GenDifS_Node():\n",
    "    def __init__(self, xml_node, *,\n",
    "                 parent_gdn = None,\n",
    "                 depth = None\n",
    "                ):\n",
    "        self.id = xml_node.get('ID')\n",
    "        self.text = xml_node.get('TEXT')\n",
    "        self.depth = depth\n",
    "        self.parent_gdn = parent_gdn\n",
    "        self.xml_node = xml_node\n",
    "        self.children_gdn_list=[]\n",
    "        self.python_type = self.__class__.__name__\n",
    "\n",
    "        self.dct, self.entities = self.string_to_dict_of_entities(self.text)\n",
    "\n",
    "        self.gendifs_type = self.gendifs_dct_type()\n",
    "\n",
    "\n",
    "        # computed by collect_context()\n",
    "        self.context = {}\n",
    "        self.context[\"id\"] = self.id\n",
    "\n",
    "        self.entity_context = {}\n",
    "        self.entity_context[\"id\"] = self.id\n",
    "\n",
    "        # computed by compile()\n",
    "        self.code = {}\n",
    "        \n",
    "        # self.molecule_per_language = None # pointer to instance of class GenDifS_molecule_per_language\n",
    "        # self.molecule_dict = {} \n",
    "\n",
    "        # self.integrity = {}\n",
    "        self.augment_dict = {}\n",
    "        \n",
    "    def __str__(self):\n",
    "        \"\"\"string representation of parsed XML Element node/@TEXT\"\"\"\n",
    "        return f\"{self.id} ({self.python_type}) {self.text}\\n {self.molecule_dict}\"\n",
    "\n",
    "    def gendifs_dct_type(self):\n",
    "        dl = list(self.dct.keys())\n",
    "        if dl[0] == \"default\": dl.pop(0)\n",
    "        return dl[0] if len(dl) >= 1 else None\n",
    "\n",
    "    def string_to_dict_of_entities(self, string): \n",
    "        \"\"\"parse a string like `ONE text one TWO some other text`\n",
    "        into a a dict like `{\"ONE\": \"text_one\", \"TWO\": \"some_other_text\" }`\n",
    "        \"\"\"\n",
    "        currect_token = \"default\" # since we accept only isupper() tokens from user input, there is no conflict\n",
    "        dct = string2dict( string )\n",
    "        if len(dct[\"default\"]) == 0:\n",
    "            dct.pop(\"default\")\n",
    "\n",
    "        dct2 = { k: \"_\".join(v) for k,v in dct.items() }\n",
    "        \n",
    "        dct3 = { k: Entity( self, \"_\".join(v) ) for k,v in dct.items() }\n",
    "        if verbose >= 3: print(f\"318: {dct3=}\")\n",
    "\n",
    "        return dct2, dct3\n",
    "\n",
    "#070\n",
    "#    def generate_molecule_per_language(self):\n",
    "#        self.molecule_per_language = GenDifS_molecule_per_language(self.context, self.entity_context)\n",
    "#        self.molecule_dict = self.molecule_per_language.compile()\n",
    "\n",
    "\n",
    "    def compile(self):\n",
    "        \n",
    "        codeclass2fn = {\n",
    "            'taxon': self.code_taxon,\n",
    "            'TAXONOMY': self.code_TAXONOMY,\n",
    "            'ISA': self.code_ISA,\n",
    "            'BY': self.code_BY_FROM, \n",
    "            'BY_SOME': self.code_BY_FROM_SOME,\n",
    "            'SUP': self.code_SUP,\n",
    "##            'ISA_SOME': self.code_ISA_SOME,\n",
    "            'UNCONDITIONAL_SOME': self.code_UNCONDITIONAL_SOME,\n",
    "            'DEF': self.code_DEF,\n",
    "#            'REL': self.code_REL\n",
    "        }\n",
    "\n",
    "        codeclass = self.context.get(\"codeclass\") # we get a string\n",
    "        if codeclass == None:\n",
    "            if verbose >= 1:\n",
    "                print(f\"WARNING 510: {self.context=}\")\n",
    "        elif codeclass not in codeclass2fn:\n",
    "            if verbose >= 1:\n",
    "                print(f\"WARNING 513: {codeclass=} not in codeclass2fn\")\n",
    "        else:\n",
    "            # call the code generating function\n",
    "            # ALT codeclass2fn[codeclass](**self.context)\n",
    "            codeclass2fn[codeclass](**self.entity_context)\n",
    "        return self.code\n",
    "\n",
    "    #\n",
    "    # the \"meat\" of the library: the functions which actualle generate code\n",
    "    #\n",
    "\n",
    "\n",
    "    def code_taxon(self, *, id, codeclass, species):\n",
    "\n",
    "        # these definitions here are redundant:\n",
    "        # most other snippets also define the involved classes\n",
    "\n",
    "        # -----------------\n",
    "        lang = \"RDFS\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\"{ species.c() } a owl:Class . \"\"\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "        \n",
    "        # lang = \"RDFS\"\n",
    "        # ttl = [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "        #         f\"\"\":{species} a owl:Class . \"\"\" ]\n",
    "        # # ttl += [ f\"\"\":{species} rdfs:label ... .\"\"\" ]\n",
    "        # self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "\n",
    "\n",
    "        # -----------------\n",
    "        lang = \"RDFStest\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\"{ species.e(id) } a { species.c() } . \"\"\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # lang = \"RDFStest\"\n",
    "        # ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "        #          f\"\"\"ex:{species}_{id} a :{species} . \"\"\" ]\n",
    "        # self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "\n",
    "        \n",
    "                \n",
    "    def code_TAXONOMY(self, *, id, codeclass, taxonomy_name, graph, species_list):\n",
    "\n",
    "        # lang = \"DE\"\n",
    "        # ttl =  [ f\"\"\"# {id} {lang}\"\"\" ]\n",
    "        # ttl += [ f\"\"\":{taxonomy_name} :explanation \"Ein Spiel in einer Phantasiewelt, es heißt *{taxonomy_name}*. Die Dinge in der Phantasiewelt haben abstrakte Bezeichner, die wir nicht verstehen, nur der Spieleleiter kennt sie. Wir beschreiben Dinge durch Attribute und ihre Werte. Wir wollen eine Taxonomie rekonstruieren. Unser abstraktes Datenmodell ist OWL-RL in der Serialisierung Turtle.\" . \"\"\" ]\n",
    "        # self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # TBD: more detailed ontology metadata etc.\n",
    "\n",
    "        # lang = \"OWL\"\n",
    "        # ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "        #          f\"\"\":{taxonomy_name} a owl:Ontology; rdfs:label \"{taxonomy_name}\" .\"\"\" ]\n",
    "        # self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # lang = \"RDFS\"\n",
    "        # ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "        #          f\"\"\":{taxonomy_name} rdfs:label \"RDFS class tree {taxonomy_name}\" .\"\"\" ]\n",
    "        # self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "\n",
    "        #ex:Ont a owl:Ontology .\n",
    "        #ex:Ont owl:imports <http://www.w3.org/TR/skos-reference/skos-owl1-dl.rdf> .\n",
    "\n",
    "        \n",
    "        # -----------------\n",
    "        lang = \"SKOS\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\"[ owl:imports <http://www.w3.org/TR/skos-reference/skos-owl1-dl.rdf> ] .\"\"\" ,\n",
    "                 f\"{taxonomy_name.e(None, prefix='cpt')} a skos:ConceptScheme .\" ]\n",
    "        ttl += [ f\"{ species.e(None, prefix='cpt') } a skos:Concept ;\\n   skos:inScheme {taxonomy_name.e(None, prefix='cpt')} .\" for species in species_list ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "        \n",
    "        # lang = \"SKOS\"\n",
    "        # ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "        #          f\"cpt:{taxonomy_name} a skos:ConceptScheme .\" ]\n",
    "        # ttl += [ f\"cpt:{species} a skos:Concept ;\\n   skos:inScheme cpt:{taxonomy_name} .\" for species in species_list ]\n",
    "        # self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        \n",
    "    def code_DEF(self, *, id, codeclass, genus, source, text_list):\n",
    "        ...\n",
    "        \n",
    "        # # class definition\n",
    "        # lang = \"DE\"\n",
    "        # ttl =  [ f\"\"\"# {id} {lang}\"\"\" ]\n",
    "        # ttl += [ f\"\"\":{genus} :explanation \"{text} ({source})\" .\"\"\" for text in text_list ]\n",
    "        # self.code[lang] = \"\\n\".join(ttl)\n",
    "        \n",
    "        # lang = \"SKOS\"\n",
    "        # ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "        #          f\"\"\"{ genus.e(None, prefix='cpt') } a skos:Concept .\"\"\" ]\n",
    "        # ttl += [ f\"\"\"{ genus.e(None, prefix='cpt') } skos:scopeNote \"{text}  ({source})\" .\"\"\" for text in text_list ]\n",
    "        # self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        \n",
    "    def code_ISA(self, *, id, codeclass, genus, species_list):\n",
    "\n",
    "        # lang = \"DE\"\n",
    "        # ttl =  [ f\"\"\"# {id} {lang}\"\"\" ]\n",
    "        # ttl += [ f\"\"\":{species} :explanation \"*{reverse(species)}* ist eine Subklasse von *{reverse(genus)}* .\" .\"\"\" for species in species_list ]\n",
    "        # self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "\n",
    "        # -----------------\n",
    "        # class definitions\n",
    "        lang = \"RDFS\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"{ genus.c()   } a owl:Class .\" ]\n",
    "        ttl += [ f\"{ species.c() } a owl:Class;\\n   rdfs:subClassOf { genus.c() } .\" for species in species_list ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        \n",
    "        # assess DISJOINT\n",
    "        # self.augment_dict[\"RDFS\"] = { \"fn\": \"DISJOINT\",  \"species_list\": species_list }\n",
    "\n",
    "\n",
    "        lang = \"RDFStest\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"{ genus.e(id)   } a { genus.c() } .\" ]\n",
    "        ttl += [ f\"{ species.e(id) } a { species.c() } .\" for species in species_list ]\n",
    "        ttl += [ f\"{ species.e(id) } gendifs:classify_similar {genus.e(id)} .\"  for species in species_list  ] # bottom up inferencing \"ISA\"\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # self.inspect_inferencing = Test_classify_similar(id, self.code[\"RDFStest\"] + self.code[\"RDFS\"] )\n",
    "\n",
    "        # -----------------\n",
    "        lang = \"SKOS\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"{ genus.e(None, prefix='cpt') } a skos:Concept .\" ]\n",
    "        ttl += [ f\"{ species.e(None, prefix='cpt') } a skos:Concept ;\\n   skos:broader { genus.e(None, prefix='cpt') } .\" for species in species_list ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        \n",
    "        \n",
    "        \n",
    "    def code_BY_FROM(self, *, id, codeclass, genus, by, frm, species_list):\n",
    "\n",
    "        # -----------------\n",
    "\n",
    "        lang=\"RDFS\"\n",
    "        ttl = [ f\"{ species.c() } a owl:Class;\\n   rdfs:subClassOf { genus.c() } .\" for species in species_list ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        restriction_technically_owlthing = f\"BY_{id}_restriction_owlthing\"\n",
    "        restriction_userfriendly_owlthing = f\"BY_{ by.e(None) }_FROM_owlthing\"\n",
    "\n",
    "        restriction_technically = f\"BY_{id}_restriction\"\n",
    "        restriction_userfriendly = f\"BY_{ by.e(None) }_FROM_{ frm.e(None) }\"\n",
    "\n",
    "        intersection_technically = f\"BY_{id}_intersection\"\n",
    "        intersection_userfriendly = f\"{restriction_userfriendly}_INTERSECT_{ genus.e(None) }\"\n",
    "                \n",
    "        lang = \"OWL\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\"{ by.c() } rdf:type owl:ObjectProperty .\"\"\" ]\n",
    "\n",
    "        ttl += [ f\"\"\":{restriction_technically_owlthing} a owl:class ;\"\"\" ,\n",
    "                 f\"\"\"   a owl:Restriction ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:label \"{restriction_userfriendly_owlthing}\" ;\"\"\" ,\n",
    "                 f\"\"\"   owl:onProperty { by.c() } ;\"\"\" ,\n",
    "                 f\"\"\"   owl:someValuesFrom owl:Thing .\"\"\" ]\n",
    "\n",
    "        ttl += [ f\"\"\":{restriction_technically} a owl:class ;\"\"\" ,\n",
    "                 f\"\"\"   a owl:Restriction ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:label \"{restriction_userfriendly}\" ;\"\"\" ,\n",
    "                 f\"\"\"   owl:onProperty { by.c() } ;\"\"\" ,\n",
    "                 f\"\"\"   owl:someValuesFrom { frm.c() } .\"\"\" ]\n",
    "\n",
    "        ttl += [ f\"\"\":{intersection_technically} a owl:class ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:label \"{intersection_userfriendly}\" ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:subClassOf { genus.c() } ;\"\"\" ,\n",
    "                 f\"\"\"   owl:intersectionOf (:{restriction_technically} { genus.c() } ) .\"\"\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # assess MODALITY\n",
    "        # collect all information which we need to assess MUST (and also SHOULD, MAY, SHOULDNOT, MUSTNOT)\n",
    "        self.augment_dict[\"MODALITY\"] = {\"restriction_technically_owlthing\": restriction_technically_owlthing,\n",
    "                                         \"restriction_technically\": restriction_technically,\n",
    "                                         \"intersection_technically\": intersection_technically,\n",
    "                                         }\n",
    "\n",
    "        # allocate example instance plus test condition\n",
    "\n",
    "        lang = \"OWLtest\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\"ex:{restriction_technically}_example a :{restriction_technically} .\"\"\" ,\n",
    "                 f\"\"\"{ genus.e(id) } a { genus.c() } ;\"\"\" ,\n",
    "                 f\"\"\"   { by.c() } { frm.e(id) } ;\"\"\" ,\n",
    "                 f\"\"\"   gendifs:classify_similar ex:{restriction_technically}_example .\"\"\" ,\n",
    "                 f\"\"\"{frm.e(id) } a { frm.c() } .\"\"\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        self.inspect_inferencing = Test_classify_similar(id, self.code[\"OWL\"] + self.code[\"OWLtest\"] )\n",
    "        \n",
    "\n",
    "        # SKOS\n",
    "\n",
    "        skos_collection_name_technically = f\"cpt:BY_{id}_intersection\"\n",
    "        skos_collection_name_userfriendly = f\"\"\"{ genus.e(None, prefix='cpt') } BY { by.e(None, prefix='cpt') } FROM { frm.e(None, prefix='cpt') }\"\"\"\n",
    "\n",
    "        lang = \"SKOS\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"{ genus.e(None, prefix='cpt') } a skos:Concept .\" ,\n",
    "                 f\"{skos_collection_name_technically} a skos:Collection ;\" ,\n",
    "                 f\"\"\"   rdfs:label \"{skos_collection_name_userfriendly}\" .\"\"\" ]\n",
    "        ttl += [ f\"{ species.e(None, prefix='cpt') } a skos:Concept ;\\n   skos:broader { genus.e(None, prefix='cpt') } .\" for species in species_list ]\n",
    "        ttl += [ f\"{skos_collection_name_technically} skos:member { species.e(None, prefix='cpt') } .\"                    for species in species_list ]\n",
    "        ttl += [ f\"{by.e(None, prefix='cpt')} a rdfs:Property ;\\n   rdfs:subPropertyOf skos:related .\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # https://www.w3.org/TR/skos-primer/#secadvanced\n",
    "        # _:b0 rdf:type skos:Collection;\n",
    "        # skos:prefLabel \"milk by source animal\"@en;\n",
    "        # skos:member ex:cowMilk;\n",
    "        # skos:member ex:goatMilk;\n",
    "        # skos:member ex:buffaloMilk.\n",
    "\n",
    "\n",
    "        \n",
    "        \n",
    "    def code_BY_FROM_SOME(self, *, id, codeclass, genus, by, frm, some, species):\n",
    "\n",
    "        # Notebook: reference-BY-FROM-SOME.html\n",
    "        \n",
    "        # ----------\n",
    "        lang = \"DE\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ]\n",
    "        ttl += [ f\"\"\"{ species.c() } :explanation \"*{ species.c(reverse = True) }* ist eine Subklasse der Klasse *{ genus.c(reverse = True) }* .\" . \"\"\"  ]\n",
    "        # ttl += [ f\"\"\"{ genus.c() } :explanation \"*{ genus.c(reverse = True) }* hat das Attribut *{reverse(by)}* .\" .  \"\"\" ]\n",
    "        ttl += [ f\"\"\"{ some.c() } :explanation \"*{reverse(some)}* ist eine Subklasse der Klasse *{reverse(frm)}* .\" .  \"\"\" ]\n",
    "        # ttl += [ f\"\"\"{ genus.c() } :explanation \"Wenn ein Exemplar der Klasse *{ genus.c(reverse = True) }* ein Attribut *{reverse(by)}* mit einem Wert aus der Klasse  *{reverse(some)}* hat: Dann ist dieses Exemplar aus *{ genus.c(reverse = True) }* auch ein Exemplar aus  *{{ species.c(reverse = True) }}*.\" .  \"\"\" ]\n",
    "        # ttl += [ f\"\"\"{ genus.c() } :explanation \"Ein *{ genus.c(reverse = True) }*, dessen Attribut *{reverse(by)}* auf ein Element aus der Klasse *{reverse(some)}* zeigt, ist ein *{{ species.c(reverse = True) }}*.\" .  \"\"\" ]\n",
    "        ttl += [ f\"\"\"{ genus.c() } :explanation \"Ein *{ genus.c(reverse = True) }*, das für das Attribut *{reverse(by)}* einen Wert aus  *{reverse(some)}* hat, ist ein *{{ species.c(reverse = True) }}*.\" .  \"\"\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "\n",
    "        lang = \"RDFS\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"{ species.c() } a owl:Class;\\n   rdfs:subClassOf { genus.c() } .\" ]\n",
    "        ttl += [ f\"\"\"{ some.c() } rdfs:subClassOf { frm.c() } .\"\"\" ] \n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        restriction_technically = f\"BY_SOME_{id}_restriction\"\n",
    "        restriction_userfriendly = f\"BY_{ by.e(None) }_FROM_{ frm.e(None) }_SOME_{some.e(None) }\"\n",
    "                \n",
    "        intersection_technically = f\"BY_SOME_{id}_intersection\"\n",
    "        intersection_userfriendly = f\"({restriction_userfriendly})_INTERSECT_{ genus.e(None) }\"\n",
    "\n",
    "        lang = \"OWL\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\"{ by.c() } rdf:type owl:ObjectProperty .\"\"\" ,\n",
    "                 f\"\"\":{restriction_technically} a owl:class ;\"\"\" ,\n",
    "                 f\"\"\"   a owl:Restriction ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:label \"{restriction_userfriendly}\" ;\"\"\" ,\n",
    "                 f\"\"\"   owl:onProperty { by.c() } ;\"\"\" ,\n",
    "                 f\"\"\"   owl:someValuesFrom { some.c() } .\"\"\" ]                                     # some\n",
    "\n",
    "        ttl += [ f\"\"\":{intersection_technically} a owl:class ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:label \"{intersection_userfriendly}\" ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:subClassOf { species.c() } ;\"\"\" ,                                     \n",
    "                 f\"\"\"   owl:intersectionOf (:{restriction_technically} { genus.c() } ) .\"\"\" ] # change e.g. to species to test Test_classify_similar\n",
    "\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # allocate example instance plus test condition\n",
    "\n",
    "        lang = \"OWLtest\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\"{ species.e(id) } a { species.c() } .\"\"\" ,\n",
    "                 f\"\"\"{ genus.e(id) }   a { genus.c() } ;\"\"\" ,\n",
    "                 f\"\"\"   { by.c() } { some.e(id) } ;\"\"\" ,\n",
    "                 f\"\"\"   gendifs:classify_similar { species.e(id) } .\"\"\" ,\n",
    "                 f\"\"\"{ some.e(id) } a { some.c() } .\"\"\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        self.inspect_inferencing = Test_classify_similar(id, self.code[\"RDFS\"] + self.code[\"OWL\"] + self.code[\"OWLtest\"] )\n",
    "\n",
    "        \n",
    "        # -----------------\n",
    "        lang = \"SKOS\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"{ genus.e(None, prefix='cpt') } a skos:Concept .\" ,\n",
    "                 f\"{ species.e(None, prefix='cpt') } a skos:Concept ;\\n   skos:broader { genus.e(None, prefix='cpt') } .\" ,\n",
    "                 f\"{ frm.e(None, prefix='cpt') } a skos:Concept .\" ,\n",
    "                 f\"{ some.e(None, prefix='cpt') } a skos:Concept ;\\n   skos:broader { frm.e(None, prefix='cpt') } .\",\n",
    "                 f\"{ by.e(None, prefix='cpt') }   a rdfs:Property ;\\n   rdfs:subPropertyOf skos:related .\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "   \n",
    "        \n",
    "    def code_UNCONDITIONAL_SOME(self, *, id, codeclass, some, species):\n",
    "\n",
    "        # ----------\n",
    "        restriction_technically = f\"SOME_{id}_restriction\"\n",
    "        restriction_userfriendly = f\"SOME_{ some.e(None) }\"\n",
    "\n",
    "        lang = \"DE\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ]\n",
    "        ttl += [ f\"\"\"{ species.c() } :explanation \"Ein beliebiges Ding, bei dem ein beliebiges Attribut einen Wert aus *{ some.c(reverse = True) }* hat, ist ein *{ species.c(reverse = True) }*.\" .  \"\"\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        lang = \"RDFS\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\"{ species.c() } a owl:Class .\"\"\" ] \n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        lang = \"OWL\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\":{restriction_technically} a owl:class ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:subClassOf { species.c() } ;\"\"\" ,\n",
    "                 f\"\"\"   a owl:Restriction ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:label \"{restriction_userfriendly}\" ;\"\"\" ,\n",
    "                 f\"\"\"   owl:onProperty owl:topObjectProperty ;\"\"\" ,\n",
    "                 f\"\"\"   owl:someValuesFrom { some.c() } .\"\"\" ]                                     # some\n",
    " \n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # allocate example instance plus test condition\n",
    "\n",
    "        lang = \"OWLtest\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\"{ species.e(id) } a { species.c() } .\"\"\" ,\n",
    "                 f\"\"\"ex:thing_{id}  a owl:Thing ;\"\"\" ,\n",
    "                 f\"\"\"   owl:topObjectProperty { some.e(id) } ;\"\"\" ,\n",
    "                 f\"\"\"   gendifs:classify_similar { species.e(id) } .\"\"\" ,\n",
    "                 f\"\"\"{ some.e(id) } a { some.c() } .\"\"\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        self.inspect_inferencing = Test_classify_similar(id, self.code[\"RDFS\"] + self.code[\"OWL\"] + self.code[\"OWLtest\"] )\n",
    "\n",
    "        \n",
    "\n",
    "    def code_SUP(self, *, id, codeclass, species, genus_list):\n",
    "        \"\"\"intersection of all genus in genus_list is subclass of secies\"\"\"\n",
    "\n",
    "        \"\"\"\n",
    "        Stallion\n",
    "           SUP\n",
    "              Horse\n",
    "              Male\n",
    "        \"\"\"\n",
    "        # bottom up inferencing:\n",
    "        # if x is a stallion, then x is a horse; and x is a male\n",
    "\n",
    "        lang = \"RDFS\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"{ species.c() } a owl:Class .\" ]\n",
    "        ttl += [ f\"{ genus.c() } a owl:Class.\\n   { species.c() } rdfs:subClassOf { genus.c() } .\" for genus in genus_list ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # top down inferencing:\n",
    "        # if x is a horse and x is a male, then x is a stallion\"\"\"\n",
    "\n",
    "        intersection_technically = f\"SUP_{id}\"\n",
    "        sup_genus_join = \" \".join([ f\"{ g.c() }\" for g in genus_list ])\n",
    "        \n",
    "        lang = \"OWL\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ,\n",
    "                 f\"\"\":{intersection_technically} a owl:class ;\"\"\" ,\n",
    "                 f\"\"\"   rdfs:subClassOf { species.c() } ;\"\"\" ,   # comment out to test self.inspect_inferencing\n",
    "                 f\"\"\"   owl:intersectionOf ({ sup_genus_join } ) .\"\"\" ]\n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        # we have one single example of species:\n",
    "        # this species example is element of all genus classes involved, but NOT element of the intersection\n",
    "        # (which in turn is a sublass of the species in question).\n",
    "\n",
    "        lang = \"OWLtest\"\n",
    "        ttl =  [ f\"\"\"# {id} {lang}\"\"\" ]\n",
    "        ttl += [ f\"\"\"{ species.e(id) } a { genus.c() } .\"\"\" for genus in genus_list ]  # species genus is correct here\n",
    "        ttl += [ f\"\"\"{ species.e(id) }_2 a { species.c() } .\"\"\" ]\n",
    "        ttl += [ f\"\"\"{ species.e(id) } gendifs:classify_similar { species.e(id) }_2 .\"\"\" ]        \n",
    "        self.code[lang] = \"\\n\".join(ttl)\n",
    "\n",
    "        self.inspect_inferencing = Test_classify_similar(id, self.code[\"OWL\"] + self.code[\"OWLtest\"] )\n",
    "\n",
    "\n",
    "    def code_ISA_SOME(self, *, id, codeclass, genus, by, some, species_list): ...\n",
    "\n",
    "    def code_REL(self, *, id, codeclass, by, some, species_list): ...\n",
    "\n",
    "\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b1d80f72",
   "metadata": {},
   "outputs": [],
   "source": [
    "\n",
    "class Entity():\n",
    "\n",
    "    def __init__(self, source_gdn, text ):\n",
    "\n",
    "        self.source_gdn = source_gdn\n",
    "        self.text = text\n",
    "        self.it_is_a_uri = len(text) > 0 and text[0] == '<' and text[-1] == '>'\n",
    "\n",
    "    def __str__(self):\n",
    "        return self.text\n",
    "\n",
    "    def __repr__(self):\n",
    "        return f\"\"\"Entity({self.source_gdn.id=}, {self.text=})\"\"\"\n",
    "\n",
    "    # create category curi\n",
    "    def c(self, *, prefix = \"\", reverse = False):\n",
    "        return self.text if self.it_is_a_uri else prefix + \":\" +  self.uri_to_string(self.text)\n",
    "\n",
    "    # create example, i.e. an instance \n",
    "    def e(self, id, *, prefix = \"ex\", reverse = False):\n",
    "        myid = self.source_gdn.id if id == None else id\n",
    "        if self.it_is_a_uri:\n",
    "            r = prefix + \":\" + self.uri_to_string(self.text[1:-1])\n",
    "            if id != None: r += \"_\" + myid\n",
    "        else:\n",
    "            r = prefix + \":\" + self.uri_to_string(self.text)\n",
    "            if id != None: r += \"_\" + myid\n",
    "        return r\n",
    "\n",
    "    def uri_to_string(self, uri):\n",
    "        return \"\".join([ c if c.isalnum() else \"_\" for c in self.text  ])\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "78560d04",
   "metadata": {},
   "outputs": [],
   "source": [
    "class GenDifS_Differentia(GenDifS_Node):\n",
    "    \n",
    "    def __init__(self, xml_node, *, parent_gdn=None, depth=None):\n",
    "        super().__init__(xml_node, parent_gdn=parent_gdn, depth=depth)\n",
    "    \n",
    "    def describe(self, *, indent = False):\n",
    "        d = self.depth if indent else 0\n",
    "        print(f\"--- {d*'  '}differentia {self.gendifs_type}: {self.dct} --- \\n\")\n",
    "        print(f\"{d*'  '}{self.context=} \\n\")\n",
    "        # print(f\"{d*'  '}{self.molecule_per_language=} \\n\")\n",
    "        print(f\"{d*'  '}{self.code.keys()=}\\n\")\n",
    "        print(f\"{d*'  '}{self.code=} \\n\")\n",
    "        print(f\"{d*'  '}{self.augment_dict=} \\n\")\n",
    "        #print(f\"{d*'  '}{(20-d)*'-'}\")\n",
    "\n",
    "    def collect_context(self):\n",
    "\n",
    "        #print(\"367:\", self.gendifs_type)\n",
    "\n",
    "        match self.gendifs_type:\n",
    "\n",
    "            case 'TAXONOMY':\n",
    "                self.context[\"codeclass\"] = \"TAXONOMY\"\n",
    "                self.context[\"taxonomy_name\"]   = self.dct.get(\"TAXONOMY\") if self.dct.get(\"TAXONOMY\") != '' else self.id\n",
    "                self.context[\"graph\"] = f\"graph_{self.id}\"\n",
    "                self.context[\"species_list\"] = [ child.dct[\"default\"] for child in self.children_gdn_list ] \n",
    "\n",
    "                self.entity_context[\"codeclass\"] = \"TAXONOMY\"\n",
    "                self.entity_context[\"taxonomy_name\"]   = self.entities.get(\"TAXONOMY\") if self.entities.get(\"TAXONOMY\") != '' else Entity(self, self.id)\n",
    "                self.entity_context[\"graph\"] = Entity(self, f\"graph_{self.id}\")\n",
    "                self.entity_context[\"species_list\"] = [ child.entities[\"default\"] for child in self.children_gdn_list ] \n",
    "\n",
    "            case 'ISA':\n",
    "                self.context[\"codeclass\"] = \"ISA\"\n",
    "                self.context[\"genus\"] = \"anything\" if self.parent_gdn == None else self.parent_gdn.dct[\"default\"]\n",
    "                self.context[\"species_list\"] = [ child.dct[\"default\"] for child in self.children_gdn_list ] \n",
    "\n",
    "                self.entity_context[\"codeclass\"] = \"ISA\"\n",
    "                self.entity_context[\"genus\"] = Entity(self, \"anything\") if self.parent_gdn == None else self.parent_gdn.entities[\"default\"]\n",
    "                self.entity_context[\"species_list\"] = [ child.entities[\"default\"] for child in self.children_gdn_list ] \n",
    "\n",
    "            case 'BY':\n",
    "                self.context[\"codeclass\"] = \"BY\"\n",
    "                self.context[\"genus\"] = \"anything\" if self.parent_gdn == None else self.parent_gdn.dct[\"default\"]\n",
    "                self.context[\"by\"]   = self.dct.get(\"BY\", \"anyproperty\")\n",
    "                self.context[\"frm\"] = self.dct.get(\"FROM\", \"anything\")\n",
    "                self.context[\"species_list\"] = [ child.dct[\"default\"] for child in self.children_gdn_list ] \n",
    "\n",
    "                self.entity_context[\"codeclass\"] = \"BY\"\n",
    "                self.entity_context[\"genus\"] = Entity(self, \"anything\") if self.parent_gdn == None else self.parent_gdn.entities[\"default\"]\n",
    "                self.entity_context[\"by\"]   = self.entities.get(\"BY\", \"anyproperty\")\n",
    "                self.entity_context[\"frm\"] = self.entities.get(\"FROM\", \"anything\")\n",
    "                self.entity_context[\"species_list\"] = [ child.entities[\"default\"] for child in self.children_gdn_list ] \n",
    "\n",
    "            case 'SUP':\n",
    "                \"\"\"SUP: species, list of genus1, genus2 (to generate intersection)\"\"\"\n",
    "                self.context[\"codeclass\"] = \"SUP\"\n",
    "                self.context[\"species\"] = \"anything\" if self.parent_gdn == None else self.parent_gdn.dct.get(\"default\", \"context_SUP_NONE_species\")\n",
    "                self.context[\"genus_list\"] = [ child.dct[\"default\"] for child in self.children_gdn_list ] \n",
    "\n",
    "                self.entity_context[\"codeclass\"] = \"SUP\"\n",
    "                self.entity_context[\"species\"] = Entity(self, \"anything\") if self.parent_gdn == None else self.parent_gdn.entities[\"default\"]\n",
    "                self.entity_context[\"genus_list\"] = [ child.entities[\"default\"] for child in self.children_gdn_list ] \n",
    "\n",
    "                \n",
    "            case 'SOME':\n",
    "                try:\n",
    "                    parent_differentia_gendifs_type = self.parent_gdn.parent_gdn.gendifs_type\n",
    "                except:\n",
    "                    parent_differentia_gendifs_type = None\n",
    "                \n",
    "                match parent_differentia_gendifs_type:\n",
    "                    case 'BY':\n",
    "                        self.context[\"codeclass\"] = \"BY_SOME\"\n",
    "                        self.context[\"genus\"] = self.parent_gdn.parent_gdn.parent_gdn.dct.get(\"default\", \"anything\")\n",
    "                        self.context[\"by\"]   = self.parent_gdn.parent_gdn.dct.get(\"BY\", \"anyproperty\")\n",
    "                        self.context[\"frm\"] = self.parent_gdn.parent_gdn.dct.get(\"FROM\", \"anything\")\n",
    "                        self.context[\"some\"] = self.dct.get(\"SOME\", \"anything\")\n",
    "                        self.context[\"species\"] = self.parent_gdn.dct.get(\"default\", \"context_BY_SOME_NONE_species\")\n",
    "\n",
    "                        self.entity_context[\"codeclass\"] = \"BY_SOME\"\n",
    "                        self.entity_context[\"genus\"] = self.parent_gdn.parent_gdn.parent_gdn.entities.get(\"default\", Entity(self, \"anything\"))\n",
    "                        self.entity_context[\"by\"]   = self.parent_gdn.parent_gdn.entities.get(\"BY\", Entity(self, \"anyproperty\"))\n",
    "                        self.entity_context[\"frm\"] = self.parent_gdn.parent_gdn.entities.get(\"FROM\", Entity(self, \"anything\"))\n",
    "                        self.entity_context[\"some\"] = self.entities.get(\"SOME\", Entity(self, \"anything\"))\n",
    "                        self.entity_context[\"species\"] = self.parent_gdn.entities.get(\"default\", Entity(self, \"context_BY_SOME_NONE_species\"))\n",
    "\n",
    "                    case 'ISA':\n",
    "                        self.context[\"codeclass\"] = \"BY_SOME\"\n",
    "                        self.context[\"genus\"] = self.parent_gdn.parent_gdn.parent_gdn.dct.get(\"default\", \"anything\")\n",
    "                        self.context[\"by\"] = \"anyproperty\"\n",
    "                        self.context[\"frm\"] = \"anything\"\n",
    "                        self.context[\"some\"] = self.dct.get(\"SOME\", \"anything\")\n",
    "                        self.context[\"species\"] = self.parent_gdn.dct.get(\"default\", \"context_ISA_SOME_NONE_species\")\n",
    "\n",
    "                        self.entity_context[\"codeclass\"] = \"BY_SOME\"\n",
    "                        self.entity_context[\"genus\"] = self.parent_gdn.parent_gdn.parent_gdn.entities.get(\"default\", Entity(self, \"anything\"))\n",
    "                        self.entity_context[\"by\"] = Entity(self, \"anyproperty\")\n",
    "                        self.entity_context[\"frm\"] = Entity(self, \"anything\")\n",
    "                        self.entity_context[\"some\"] = self.entities.get(\"SOME\", Entity(self, \"anything\"))\n",
    "                        self.entity_context[\"species\"] = self.parent_gdn.entities.get(\"default\", Entity(self, \"context_ISA_SOME_NONE_species\"))\n",
    "\n",
    "                    case _:\n",
    "                        self.context[\"codeclass\"] = \"UNCONDITIONAL_SOME\"\n",
    "                        self.context[\"some\"] = self.dct.get(\"SOME\", \"anything\")\n",
    "                        self.context[\"species\"] = self.parent_gdn.dct.get(\"default\", \"context_MINIMAL_SOME_NONE_species\")\n",
    "\n",
    "                        self.entity_context[\"codeclass\"] = \"UNCONDITIONAL_SOME\"\n",
    "                        self.entity_context[\"some\"] = self.entities.get(\"SOME\", Entity(self, \"anything\"))\n",
    "                        self.entity_context[\"species\"] = self.parent_gdn.entities.get(\"default\", Entity(self, \"context_MINIMAL_SOME_NONE_species\"))\n",
    "\n",
    "            case 'REL':\n",
    "                ...\n",
    "\n",
    "            case 'DEF':\n",
    "                self.context[\"codeclass\"] = \"DEF\"\n",
    "                self.context[\"genus\"] = self.parent_gdn.dct[\"default\"]\n",
    "                self.context[\"source\"] = self.dct.get(\"DEF\", \"this_taxonomy\")\n",
    "                self.context[\"text_list\"] = [ child.text for child in self.children_gdn_list ] \n",
    "\n",
    "                self.entity_context[\"codeclass\"] = \"DEF\"\n",
    "                self.entity_context[\"genus\"] = self.parent_gdn.entities[\"default\"]\n",
    "                self.entity_context[\"source\"] = self.entities.get(\"DEF\", Entity(self, \"this_taxonomy\"))\n",
    "                self.entity_context[\"text_list\"] = [ child.text for child in self.children_gdn_list ] \n",
    "\n",
    "            case _:\n",
    "                if verbose >= 1:\n",
    "                    print(f'Error 297: collect_context(): unknown gendifs_type {self.gendifs_type} in node \"{self.text}\"') \n",
    "\n",
    "\n",
    "    def augment_genus_species_DEPRECATED(self, augment_graph):\n",
    "\n",
    "        def instances(graph, category):\n",
    "            q = f\"\"\"SELECT ?i WHERE {{ ?i a {category} . }}\"\"\"\n",
    "            return { r[0].n3() for r in graph.query(q) }\n",
    "        \n",
    "        if not( \"genus\" in self.entity_context and \"species_list\" in self.entity_context):\n",
    "            return None\n",
    "        \n",
    "        genus_instances = instances(augment_graph, self.entity_context[\"genus\"].c())\n",
    "\n",
    "        # for each species: all instances\n",
    "        species_instances_dict = {}\n",
    "        for s in self.entity_context[\"species_list\"]:\n",
    "            instance_set = instances( augment_graph, s.c() )\n",
    "            if s.c() not in species_instances_dict:\n",
    "                species_instances_dict[ s.c() ] = set()\n",
    "            species_instances_dict[ s.c() ].update(instance_set)\n",
    "        if verbose >= 3:\n",
    "            print(f\"557 {species_instances_dict=}\")\n",
    "\n",
    "\n",
    "        # edge: borderline between data lake and land, coming from direction lake\n",
    "        union_of_species_instances = set()\n",
    "        for s in species_instances_dict.values():\n",
    "            #print(\"562\", s)\n",
    "            union_of_species_instances.update( s )\n",
    "        edge = genus_instances - union_of_species_instances\n",
    "        if verbose >= 2:\n",
    "            print(f\"546: {self.id=}, {edge=})\")\n",
    "        self.augment_set[\"edge\"] = edge\n",
    "\n",
    "        # DISJOINT\n",
    "        non_disjoint = {}\n",
    "        k = list( species_instances_dict.keys() )\n",
    "        for i in range(len(k)):\n",
    "            for j in range(i+1, len(k)):\n",
    "                isct = species_instances_dict[ k[i] ].intersection(species_instances_dict[ k[j] ])\n",
    "                if len(isct) > 0:\n",
    "                    non_disjoint[ (k[i], k[j]) ] = isct\n",
    "        self.augment_set[\"non_disjoint\"] = non_disjoint\n",
    "\n",
    "                \n",
    "        "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5d5d96e3",
   "metadata": {},
   "outputs": [],
   "source": [
    "class GenDifS_Taxon(GenDifS_Node):\n",
    "    \n",
    "    def __init__(self, xml_node, *, parent_gdn=None, depth=None):\n",
    "        super().__init__(xml_node, parent_gdn=parent_gdn, depth=depth)\n",
    "\n",
    "    def describe(self, *, indent = False):\n",
    "        d = self.depth if indent else 0\n",
    "        print(f\"{d*'  '}{self.dct.get('default')}\")\n",
    "        print(f\"{d*'  '}{self.context=}\")\n",
    "        print(f\"{d*'  '}{(20-d)*'='}\")\n",
    "\n",
    "    def collect_context(self):\n",
    "        self.context[\"codeclass\"] = \"taxon\"\n",
    "        self.context[\"species\"] = self.dct.get(\"default\")\n",
    "\n",
    "        self.entity_context[\"codeclass\"] = \"taxon\"\n",
    "        self.entity_context[\"species\"] = self.entities.get(\"default\")\n",
    "\n",
    "    def augment_genus_species_DEPRECATED(self, augment_graph):\n",
    "        ..."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f2e7051e",
   "metadata": {},
   "source": [
    "# generate ttl code\n",
    "\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "59dd0335",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "class Test_classify_similar:\n",
    "        def __init__(self, id, ttl, *, source = None ):\n",
    "        \n",
    "            self.ns = \"\"\"\n",
    "            PREFIX : <urn:default#>\n",
    "            PREFIX ex: <urn:ex#>\n",
    "            PREFIX cpt: <urn:cpt#>\n",
    "            PREFIX gendifs: <urn:gendifs#>\n",
    "\n",
    "            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>\n",
    "            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>\n",
    "            PREFIX owl: <http://www.w3.org/2002/07/owl#>\n",
    "            \"\"\"\n",
    "\n",
    "            self.test_errors = None\n",
    "            self.source = source\n",
    "            self.ttl = ttl\n",
    "\n",
    "            self.test_results = {}\n",
    "            \n",
    "            self.g1 = rdflib.Graph()\n",
    "            #print(f\"\"\"684: =====\\n{self.ns + self.ttl}\"\"\")\n",
    "            self.g1.parse(data = self.ns + self.ttl)\n",
    "\n",
    "            categories = {}\n",
    "            \n",
    "            # classes of example1 before inferencing\n",
    "\n",
    "            q1 = \"\"\"\n",
    "            SELECT ?example1 ?example2 ?category1\n",
    "            WHERE {\n",
    "            ?example1 gendifs:classify_similar ?example2 .\n",
    "            ?example1 a ?category1 .\n",
    "            }\"\"\"\n",
    "\n",
    "            c = \"cat1_before_inferencing\"\n",
    "\n",
    "            for example1, example2, category1 in self.g1.query(q1):  # g1, q1, cat1\n",
    "                ex_1 = example1.encode()\n",
    "                ex_2 = example2.encode()\n",
    "                cat_1 = category1.encode()\n",
    "\n",
    "                if (ex_1, ex_2, c) not in categories:\n",
    "                    categories[(ex_1, ex_2, c)] = set()\n",
    "                categories[(ex_1, ex_2, c)].add(cat_1)\n",
    "\n",
    "            # classes of example2 before inferencing\n",
    "            \n",
    "            q2 = \"\"\"\n",
    "            SELECT ?example1 ?example2 ?category2\n",
    "            WHERE {\n",
    "            ?example1 gendifs:classify_similar ?example2 .\n",
    "            ?example2 a ?category2 .\n",
    "            }\"\"\"\n",
    "\n",
    "            c = \"cat2_before_inferencing\"\n",
    "            for example1, example2, category2 in self.g1.query(q2):  # g1, q2, cat2\n",
    "                ex_1 = example1.encode()\n",
    "                ex_2 = example2.encode()\n",
    "                cat_2 = category2.encode()\n",
    "\n",
    "                if (ex_1, ex_2, c) not in categories:\n",
    "                    categories[(ex_1, ex_2, c)] = set()\n",
    "                categories[(ex_1, ex_2, c)].add(cat_2)\n",
    "\n",
    "                \n",
    "            self.g2 = rdflib.Graph()\n",
    "            self.g2.parse(data = self.ns + self.ttl)\n",
    "            owlrl.DeductiveClosure(owlrl.OWLRL_Semantics, axiomatic_triples = False).expand(self.g2)\n",
    "            \n",
    "            # Die Instanz `ex_1` ist nach dem Inferencing eine Instanz potentiell vieler Klassen `cat_1`.\n",
    "            # (Das gilt übrigens auch für `ex_2`.)\n",
    "            # Eine dieser vielen Klassen muss auch die (idealerweise einzige) Klasse `cat_2` sein,\n",
    "            # in der `ex_2` schon *vor* dem Inferencing enthalten ist.\n",
    "\n",
    "            c = \"cat1_after_inferencing\"\n",
    "            for example1, example2, category1 in self.g2.query(q1):  # g2, q1, cat1\n",
    "                ex_1 = example1.encode()\n",
    "                ex_2 = example2.encode()\n",
    "                cat_1 = category1.encode()\n",
    "\n",
    "                if (ex_1, ex_2, c) not in categories:\n",
    "                    categories[(ex_1, ex_2, c)] = set()\n",
    "                categories[(ex_1, ex_2, c)].add(cat_1)\n",
    "\n",
    "            d2 = {}\n",
    "            for (x,y,z), v in categories.items():\n",
    "                if (x,y) not in d2:\n",
    "                    d2[(x,y)] = {}\n",
    "                if z not in d2[(x,y)]:\n",
    "                    d2[(x,y)][z] = {}\n",
    "                d2[(x,y)][z] = v\n",
    "\n",
    "            for tuple, sets in d2.items():\n",
    "                #debug\n",
    "                #print(f\"\\n{tuple}:\")\n",
    "                #for k,v in sets.items():\n",
    "                #    print(f\"   {k}: {v}\")\n",
    "\n",
    "                self.test_results[tuple] = {}\n",
    "                # are cat1 and cat2 not empty, are they instances of some classes?\n",
    "                self.test_results[tuple][\"cat1 not empty\"] = sets[\"cat1_before_inferencing\"] != set()\n",
    "                self.test_results[tuple][\"cat2 not empty\"] = sets[\"cat2_before_inferencing\"] != set()\n",
    "\n",
    "                # are cat1_before_inferencing and cat2_before_inferencing disjoint?\n",
    "                self.test_results[tuple][\"disjoint before\"] = sets[\"cat1_before_inferencing\"].intersection(sets[\"cat2_before_inferencing\"]) == set() \n",
    "                \n",
    "                # has cat2_before_inferencing become a subset of von cat1_after_inferencing?\n",
    "                self.test_results[tuple][\"subset after\"] = sets[\"cat2_before_inferencing\"].intersection(sets[\"cat1_after_inferencing\"]) == sets[\"cat2_before_inferencing\"]\n",
    "\n",
    "                self.test_results[tuple][\"ok\"] = all( [ boolean for boolean in self.test_results[tuple].values() ] )\n",
    "\n",
    "                if self.test_results[tuple][\"ok\"]:\n",
    "                    if verbose >= 3:\n",
    "                        print(f\"\"\"766: Test_classify_similar OK, {id=}, {tuple=}.\"\"\")\n",
    "                else:\n",
    "                    if verbose >= 1:\n",
    "                        print(f\"\"\"WARNING 767: Test_classify_similar FAILED, {id=},\\n   {self.test_results=}\"\"\")\n",
    "                        print(\"-- self.g1 --\")\n",
    "                        print(self.g1.serialize())\n",
    "                        print(\"-- self.g2 --\")\n",
    "                        print(self.g2.serialize())\n",
    "                        print(\"-------\")\n",
    "\n",
    "            self.test_ok = all( [ v[\"ok\"] for v in self.test_results.values() ] )"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "135589a4",
   "metadata": {},
   "outputs": [],
   "source": [
    "class Integrity_Assessment:\n",
    "\n",
    "    def __init__(self, g1, taxonomy, *, inference = True, verbose = 1):\n",
    "        self.taxonomy = taxonomy\n",
    "        self.assess_dict = {}\n",
    "\n",
    "        # g1: original graph; no other classes etc., i.e. no taxonomy added\n",
    "        self.g1 = g1\n",
    "        self.verbose = verbose\n",
    "\n",
    "        # create deep copy of g1\n",
    "        self.g2 = rdflib.Graph().parse(data = self.g1.serialize() )\n",
    "\n",
    "        # add taxonomy t-box to g2\n",
    "        for language in [ \"RDFS\", \"OWL\" ]:\n",
    "            if language in taxonomy.rdf_graphs:\n",
    "                self.g2 += taxonomy.rdf_graphs[language]\n",
    "\n",
    "        # what can be inferred? compute deductive closure!\n",
    "        if inference: # default case\n",
    "            owlrl.DeductiveClosure(owlrl.OWLRL_Semantics, axiomatic_triples = False).expand(self.g2)\n",
    "        else:\n",
    "            ... # do nothing, i.e. leave g2 untouched\n",
    "            \n",
    "        if self.verbose >= 2:\n",
    "            print(f\"INFO 1323: {inference=}, {len(self.g1)=}, {len(self.g2)=}\")\n",
    "\n",
    "        # perform integrity test\n",
    "        for gdn in taxonomy.dict_of_all_gdn.values():\n",
    "            self.test_INTEGRITY(gdn)\n",
    "\n",
    "        # collect ttl into the result graph\n",
    "        self.assess_graph = rdflib.Graph()\n",
    "        for gdn_info in self.assess_dict.values():\n",
    "            for integrity_check in gdn_info.values():\n",
    "                if len( integrity_check ) > 0:\n",
    "                    try:\n",
    "                        self.assess_graph.parse( data = self.taxonomy.ttl_namespaces + integrity_check )\n",
    "                    except:\n",
    "                        if verbose >= 1:\n",
    "                            print(f\"ERROR 1337: RDFLIB complains about { integrity_check }\")\n",
    "\n",
    "            \n",
    "    def test_INTEGRITY(self, gdn):\n",
    "\n",
    "        if gdn.entity_context['codeclass'] == 'TAXONOMY':\n",
    "            # we do not have integrity checks for taxonomy nodes\n",
    "            return\n",
    "                            \n",
    "        if 'species_list' not in gdn.entity_context:\n",
    "            # we do have a leaf node, nothing to do\n",
    "            return None\n",
    "\n",
    "        self.assess_dict[gdn.id] = {}\n",
    "\n",
    "        q = {} # queries\n",
    "        species_entities = gdn.entity_context['species_list']\n",
    "        genus_entity = gdn.entity_context['genus']\n",
    "\n",
    "        # instances of the genus\n",
    "        genus_instances = {  row[0].n3() for row in self.g2.query(\n",
    "            f\"\"\"SELECT ?example WHERE {{?example a {genus_entity.c()} .}}\"\"\") }\n",
    "\n",
    "        # instances of all species\n",
    "        species_instances = {}\n",
    "        for category in species_entities:\n",
    "            q[category.text]  = f\"\"\"SELECT ?example WHERE {{?example a {category.c()} .}}\"\"\"\n",
    "            species_instances[category.text] = {  row[0].n3() for row in self.g2.query(q[category.text]) }\n",
    "\n",
    "            \n",
    "        # EDGE\n",
    "        # edge: instances of genus, which are not an instance of at least one subclass (if a subclass is given)\n",
    "        # TBD: better call it the RESIDUUM?\n",
    "        union_of_species_instances = set()\n",
    "        for s in species_instances.values():\n",
    "            #print(\"1245\", s)\n",
    "            union_of_species_instances.update( s )\n",
    "        edge = genus_instances - union_of_species_instances\n",
    "        if verbose >= 2:\n",
    "            print(f\"1249: {self.id=}, {edge=})\")\n",
    "\n",
    "        edge_name = f\"\"\":EDGE_{gdn.entity_context['id']}\"\"\"\n",
    "        ttl = [ f\"\"\"{edge_name} rdfs:subClassOf :EDGE ;\\n   rdfs:label \"EDGE of {genus_entity.c()}\"  .\"\"\" ] # ;\\n   rdfs:subClassOf {genus_entity.c()}\n",
    "        ttl.extend( [ f\"\"\"{e} a {edge_name} .\"\"\" for e in edge ] )\n",
    "        self.assess_dict[gdn.id]['EDGE'] = \"\\n\".join( ttl )\n",
    "\n",
    "        \n",
    "        # DISJOINT\n",
    "        if \"DISJOINT\" in gdn.dct:\n",
    "            ttl = []\n",
    "            len_species_entities = len(species_entities)\n",
    "            for i in range( len_species_entities ):\n",
    "                for j in range( i+1, len_species_entities ):\n",
    "                    ci = species_entities[i].text\n",
    "                    cj = species_entities[j].text\n",
    "                    cij_set = species_instances[ ci ] & species_instances[ cj ] \n",
    "                    if len(cij_set) > 0:\n",
    "                        cij = f\"\"\":INTEGRITY_{gdn.entity_context['id']}_NOT_DISJOINT_{species_entities[i].source_gdn.id}_AND_{species_entities[j].source_gdn.id}\"\"\"\n",
    "                        ttl.extend( [ f\"\"\"{cij} rdfs:label \"{ci}_AND_{cj}\";\\n rdfs:subClassOf :INTEGRITY_NOT_DISJOINT . \"\"\" ] ) # , :{ci}, :{cj}\n",
    "                        for example in cij_set:\n",
    "                            ttl.append( f\"{example} a {cij} .\")\n",
    "\n",
    "            self.assess_dict[gdn.id]['DISJOINT'] = \"\\n\".join( ttl )\n",
    "            #disjoint = \"\\n\".join( ttl )\n",
    "            #print(f\"1360{disjoint}\")\n",
    "\n",
    "        \n",
    "        # MODALITY\n",
    "\n",
    "        if \"MODALITY\" in gdn.dct:\n",
    "            \n",
    "            if \"MODALITY\" not in gdn.augment_dict:\n",
    "                print(f\"ERROR 1311: MODALITY not yet implemented, {gdn.dct=}\")\n",
    "                return\n",
    "\n",
    "            \n",
    "            \n",
    "            modalities = gdn.dct[\"MODALITY\"].split(\" \")\n",
    "            #print(f\"1315: {modalities=}\")\n",
    "                \n",
    "            restriction_instances = {}\n",
    "            for restriction_name, restriction_class in gdn.augment_dict[\"MODALITY\"].items():\n",
    "                q = f\"\"\"SELECT ?example WHERE {{?example a :{restriction_class} .}}\"\"\"\n",
    "                restriction_instances[restriction_name] = {  row[0].n3() for row in self.g2.query(q) }\n",
    "\n",
    "            #print(f\"\\nXXX INFO 1325: instances of restrictions:\\n{restriction_instances}\")\n",
    "\n",
    "            \n",
    "            \n",
    "            ttl = []\n",
    "            \n",
    "            # iv ... integrity violation\n",
    "\n",
    "            # MODALITY_MUST\n",
    "            # * genus minus restriction_technically\n",
    "            \n",
    "            # iv = f\":MODALITY_{gdn.entity_context['id']}_MUST\"\n",
    "            # iv_set = genus_instances - restriction_instances['restriction_technically']\n",
    "            \n",
    "            # ttl += [ f\"\"\"{iv} rdfs:subClassOf :MODALITY . \"\"\" ]\n",
    "            # ttl += [ f\"\"\"{i} a {iv} .\"\"\" for i in iv_set ]\n",
    "            \n",
    "            # MODALITY_MUST_property_not_given_at_all\n",
    "            # * {} != genus intersect restriction_by_owlthing (thus FROM == owl:thing)\n",
    "\n",
    "            iv = f\":MODALITY_{gdn.entity_context['id']}_MUST_property_not_given_at_all\"\n",
    "            iv_set = genus_instances - restriction_instances['restriction_technically_owlthing']\n",
    "            \n",
    "            ttl += [ f\"\"\"{iv} rdfs:subClassOf :MODALITY . \"\"\" ]\n",
    "            ttl += [ f\"\"\"{i} a {iv} .\"\"\" for i in iv_set ]\n",
    "            self.assess_dict[gdn.id][\"MODALITY\"] = \"\\n\".join( ttl )\n",
    "\n",
    "            \n",
    "            # MODALITY_MUST_property_given_but_wrong_range: \n",
    "            # * {} != genus intersect ( restriction_by_owlthing minus restriction_by_from )\n",
    "            \n",
    "            iv = f\":MODALITY_{gdn.entity_context['id']}_MUST_property_given_but_wrong_range\"\n",
    "            iv_set = genus_instances & ( restriction_instances['restriction_technically_owlthing'] - restriction_instances['restriction_technically'] )\n",
    "            \n",
    "            ttl += [ f\"\"\"{iv} rdfs:subClassOf :MODALITY . \"\"\" ]\n",
    "            ttl += [ f\"\"\"{i} a {iv} .\"\"\" for i in iv_set ]\n",
    "                \n",
    "            # MODALITY_MUSTNOT_property_not_allowed\n",
    "            # *  restriction_owlthing (also FROM = owl:thing) intersect genus\n",
    "\n",
    "            \n",
    "            self.assess_dict[gdn.id][\"MODALITY\"] = \"\\n\".join( ttl )\n",
    "\n",
    "\n",
    "            "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "684114bf",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "outputs": [],
   "source": [
    "def reverse(x, *, rev = True) -> str:\n",
    "    return x\n",
    "    # if rev and x[0].isupper():\n",
    "    #     ret = x[::-1].title()\n",
    "    # elif rev:\n",
    "    #     ret =  x[::-1].lower()\n",
    "    # else:\n",
    "    #     ret = x\n",
    "    # return ret\n",
    "\n",
    "\n",
    "def focus(focus_curie_list, ttl): # ll ... 'RDF_RDFtest_OWL_OWLtest'\n",
    "    result = \"\\n\\n\".join( [ paragraph for paragraph in ttl.split(\"\\n\\n\") \\\n",
    "                         if any( [ focus_curie in paragraph for focus_curie in focus_curie_list ] ) ] )\n",
    "    return result\n",
    "\n",
    "\n",
    "def string2dict(string):\n",
    "    currect_token = \"default\" \n",
    "    dct = { currect_token: [] }\n",
    "    for token in string.split():\n",
    "        if token.isupper():\n",
    "            currect_token = token\n",
    "            if currect_token in dct:\n",
    "                if not \"warning\" in dct:\n",
    "                    dct[\"warning\"] = []\n",
    "                dct[\"warning\"].append(f\"warning from nodetext_to_dict: multiple facets {currect_token}\")\n",
    "            else:\n",
    "                dct[currect_token] = []\n",
    "        else:\n",
    "            dct[currect_token].append(token)\n",
    "    return dct"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "formats": "ipynb,py:percent"
  },
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.4"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
